@@ -28,6 +28,7 @@ from io import BytesIO
 from string import hexdigits
 import tarfile
 from time import time
+import warnings
 
 # Import from pygit2
 from ._pygit2 import Repository as _Repository, init_file_backend
@@ -39,6 +40,7 @@ from ._pygit2 import GIT_REF_SYMBOLIC
 from ._pygit2 import Reference, Tree, Commit, Blob
 from ._pygit2 import InvalidSpecError
 
+from .callbacks import git_fetch_options
 from .config import Config
 from .errors import check_error
 from .ffi import ffi, C
@@ -84,7 +86,52 @@ class BaseRepository(_Repository):
     def __iter__(self):
         return iter(self.odb)
 
+    def add_submodule(self, url, path, link=True, callbacks=None):
+        """Add a submodule to the index.
+
+        Returns: the submodule that was added.
+
+        Parameters:
+
+        url
+            The URL of the submdoule.
+
+        path
+            The path within the parent repository to add the submodule
+
+        link
+            Should workdir contain a gitlink to the repo in .git/modules vs. repo directly in workdir.
+        """
+        csub = ffi.new('git_submodule **')
+        curl = ffi.new('char[]', to_bytes(url))
+        cpath = ffi.new('char[]', to_bytes(path))
+        gitlink = 1 if link else 0
+
+        err = C.git_submodule_add_setup(csub, self._repo, curl, cpath, gitlink)
+        check_error(err)
+
+        submodule_instance = Submodule._from_c(self, csub[0])
+
+        # prepare options
+        opts = ffi.new('git_submodule_update_options *')
+        C.git_submodule_update_init_options(opts, C.GIT_SUBMODULE_UPDATE_OPTIONS_VERSION)
+
+        with git_fetch_options(callbacks, opts=opts.fetch_opts) as payload:
+            crepo = ffi.new('git_repository **')
+            err = C.git_submodule_clone(crepo, submodule_instance._subm, opts)
+            payload.check_error(err)
+
+        # clean-up the submodule repository
+        Repository._from_c(crepo[0], True)
+
+        err = C.git_submodule_add_finalize(submodule_instance._subm)
+        check_error(err)
+        return submodule_instance
+
     def lookup_submodule(self, path):
+        """
+        Lookup submodule information by name or path.
+        """
         csub = ffi.new('git_submodule **')
         cpath = ffi.new('char[]', to_bytes(path))
 
@@ -93,28 +140,26 @@ class BaseRepository(_Repository):
         return Submodule._from_c(self, csub[0])
 
     def update_submodules(self, submodules=None, init=False, callbacks=None):
+        """
+        Update a submodule. This will clone a missing submodule and checkout
+        the subrepository to the commit specified in the index of the
+        containing repository. If the submodule repository doesn't contain the
+        target commit (e.g. because fetchRecurseSubmodules isn't set), then the
+        submodule is fetched using the fetch options supplied in options.
+        """
         if submodules is None:
             submodules = self.listall_submodules()
 
         # prepare options
         opts = ffi.new('git_submodule_update_options *')
-        C.git_submodule_update_init_options(
-            opts,
-            C.GIT_SUBMODULE_UPDATE_OPTIONS_VERSION)
-        if callbacks is not None:
-            callbacks._fill_fetch_options(opts.fetch_opts)
-        if init:
-            i = 1
-        else:
-            i = 0
+        C.git_submodule_update_init_options(opts, C.GIT_SUBMODULE_UPDATE_OPTIONS_VERSION)
 
-        for submodule in submodules:
-            submodule_instance = self.lookup_submodule(submodule)
-            err = C.git_submodule_update(
-                submodule_instance._subm,
-                i,
-                opts)
-            check_error(err)
+        with git_fetch_options(callbacks, opts=opts.fetch_opts) as payload:
+            i = 1 if init else 0
+            for submodule in submodules:
+                submodule_instance = self.lookup_submodule(submodule)
+                err = C.git_submodule_update(submodule_instance._subm, i, opts)
+                payload.check_error(err)
 
         return None
 
@@ -141,10 +186,7 @@ class BaseRepository(_Repository):
     # Remotes
     #
     def create_remote(self, name, url):
-        """Create a new remote. Return a <Remote> object.
-
-        This method is deprecated, please use Remote.remotes.create()
-        """
+        warnings.warn("Use repo.remotes.create(..)", DeprecationWarning)
         return self.remotes.create(name, url)
 
     #
@@ -379,7 +421,7 @@ class BaseRepository(_Repository):
             return None
 
         # If it's a string, then it has to be valid revspec
-        if isinstance(obj, str):
+        if isinstance(obj, str) or isinstance(obj, bytes):
             obj = self.revparse_single(obj)
         elif isinstance(obj, Oid):
             obj = self[obj]
@@ -569,32 +611,81 @@ class BaseRepository(_Repository):
     #
     # Merging
     #
+    _FAVOR_TO_ENUM = {
+        'normal': C.GIT_MERGE_FILE_FAVOR_NORMAL,
+        'ours': C.GIT_MERGE_FILE_FAVOR_OURS,
+        'theirs': C.GIT_MERGE_FILE_FAVOR_THEIRS,
+        'union': C.GIT_MERGE_FILE_FAVOR_UNION,
+    }
 
-    @staticmethod
-    def _merge_options(favor):
+    _MERGE_FLAG_TO_ENUM = {
+        'find_renames': C.GIT_MERGE_FIND_RENAMES,
+        'fail_on_conflict': C.GIT_MERGE_FAIL_ON_CONFLICT,
+        'skip_reuc': C.GIT_MERGE_SKIP_REUC,
+        'no_recursive': C.GIT_MERGE_NO_RECURSIVE,
+    }
+
+    _MERGE_FLAG_DEFAULTS = {
+        'find_renames': True,
+    }
+
+    _MERGE_FILE_FLAG_TO_ENUM = {
+        'standard_style': C.GIT_MERGE_FILE_STYLE_MERGE,
+        'diff3_style': C.GIT_MERGE_FILE_STYLE_DIFF3,
+        'simplify_alnum': C.GIT_MERGE_FILE_SIMPLIFY_ALNUM,
+        'ignore_whitespace': C.GIT_MERGE_FILE_IGNORE_WHITESPACE,
+        'ignore_whitespace_change': C.GIT_MERGE_FILE_IGNORE_WHITESPACE_CHANGE,
+        'ignore_whitespace_eol': C.GIT_MERGE_FILE_IGNORE_WHITESPACE_EOL,
+        'patience': C.GIT_MERGE_FILE_DIFF_PATIENCE,
+        'minimal': C.GIT_MERGE_FILE_DIFF_MINIMAL,
+    }
+
+    _MERGE_FILE_FLAG_DEFAULTS = {}
+
+    @classmethod
+    def _flag_dict_to_bitmask(cls, flag_dict, flag_defaults, mapping, label):
+        """
+        Converts a dict eg {"find_renames": True, "skip_reuc": True} to
+        a bitmask eg C.GIT_MERGE_FIND_RENAMES | C.GIT_MERGE_SKIP_REUC.
+        """
+        merged_dict = {**flag_defaults, **flag_dict}
+        bitmask = 0
+        for k, v in merged_dict.items():
+            enum = mapping.get(k, None)
+            if enum is None:
+                raise ValueError("unknown %s: %s" % (label, k))
+            if v:
+                bitmask |= enum
+        return bitmask
+
+    @classmethod
+    def _merge_options(cls, favor='normal', flags={}, file_flags={}):
         """Return a 'git_merge_opts *'"""
 
-        def favor_to_enum(favor):
-            if favor == 'normal':
-                return C.GIT_MERGE_FILE_FAVOR_NORMAL
-            elif favor == 'ours':
-                return C.GIT_MERGE_FILE_FAVOR_OURS
-            elif favor == 'theirs':
-                return C.GIT_MERGE_FILE_FAVOR_THEIRS
-            elif favor == 'union':
-                return C.GIT_MERGE_FILE_FAVOR_UNION
-            else:
-                return None
-
-        favor_val = favor_to_enum(favor)
+        favor_val = cls._FAVOR_TO_ENUM.get(favor, None)
         if favor_val is None:
-            raise ValueError("unkown favor value %s" % favor)
+            raise ValueError("unknown favor: %s" % favor)
+
+        flags_bitmask = Repository._flag_dict_to_bitmask(
+            flags,
+            cls._MERGE_FLAG_DEFAULTS,
+            cls._MERGE_FLAG_TO_ENUM,
+            "merge flag"
+        )
+        file_flags_bitmask = cls._flag_dict_to_bitmask(
+            file_flags,
+            cls._MERGE_FILE_FLAG_DEFAULTS,
+            cls._MERGE_FILE_FLAG_TO_ENUM,
+            "merge file_flag"
+        )
 
         opts = ffi.new('git_merge_options *')
         err = C.git_merge_init_options(opts, C.GIT_MERGE_OPTIONS_VERSION)
         check_error(err)
 
         opts.file_favor = favor_val
+        opts.flags = flags_bitmask
+        opts.file_flags = file_flags_bitmask
 
         return opts
 
@@ -631,7 +722,7 @@ class BaseRepository(_Repository):
 
         return ret
 
-    def merge_commits(self, ours, theirs, favor='normal'):
+    def merge_commits(self, ours, theirs, favor='normal', flags={}, file_flags={}):
         """
         Merge two arbitrary commits.
 
@@ -655,8 +746,34 @@ class BaseRepository(_Repository):
 
             For all but NORMAL, the index will not record a conflict.
 
-        Both "ours" and "theirs" can be any object which peels to a commit or the id
-        (string or Oid) of an object which peels to a commit.
+        flags
+            A dict of str: bool to turn on or off functionality while merging.
+            If a key is not present, the default will be used. The keys are:
+
+            * find_renames. Detect file renames. Defaults to True.
+            * fail_on_conflict. If a conflict occurs, exit immediately instead
+              of attempting to continue resolving conflicts.
+            * skip_reuc. Do not write the REUC extension on the generated index.
+            * no_recursive. If the commits being merged have multiple merge
+              bases, do not build a recursive merge base (by merging the
+              multiple merge bases), instead simply use the first base.
+
+        file_flags
+            A dict of str: bool to turn on or off functionality while merging.
+            If a key is not present, the default will be used. The keys are:
+
+            * standard_style. Create standard conflicted merge files.
+            * diff3_style. Create diff3-style file.
+            * simplify_alnum. Condense non-alphanumeric regions for simplified
+              diff file.
+            * ignore_whitespace. Ignore all whitespace.
+            * ignore_whitespace_change. Ignore changes in amount of whitespace.
+            * ignore_whitespace_eol. Ignore whitespace at end of line.
+            * patience. Use the "patience diff" algorithm
+            * minimal. Take extra time to find minimal diff
+
+        Both "ours" and "theirs" can be any object which peels to a commit or
+        the id (string or Oid) of an object which peels to a commit.
         """
 
         ours_ptr = ffi.new('git_commit **')
@@ -671,7 +788,7 @@ class BaseRepository(_Repository):
         ours = ours.peel(Commit)
         theirs = theirs.peel(Commit)
 
-        opts = self._merge_options(favor)
+        opts = self._merge_options(favor, flags, file_flags)
 
         ffi.buffer(ours_ptr)[:] = ours._pointer[:]
         ffi.buffer(theirs_ptr)[:] = theirs._pointer[:]
@@ -681,7 +798,7 @@ class BaseRepository(_Repository):
 
         return Index.from_c(self, cindex)
 
-    def merge_trees(self, ancestor, ours, theirs, favor='normal'):
+    def merge_trees(self, ancestor, ours, theirs, favor='normal', flags={}, file_flags={}):
         """
         Merge two trees.
 
@@ -707,6 +824,32 @@ class BaseRepository(_Repository):
             * union. Unique lines from each side will be used.
 
             For all but NORMAL, the index will not record a conflict.
+
+        flags
+            A dict of str: bool to turn on or off functionality while merging.
+            If a key is not present, the default will be used. The keys are:
+
+            * find_renames. Detect file renames. Defaults to True.
+            * fail_on_conflict. If a conflict occurs, exit immediately instead
+              of attempting to continue resolving conflicts.
+            * skip_reuc. Do not write the REUC extension on the generated index.
+            * no_recursive. If the commits being merged have multiple merge
+              bases, do not build a recursive merge base (by merging the
+              multiple merge bases), instead simply use the first base.
+
+        file_flags
+            A dict of str: bool to turn on or off functionality while merging.
+            If a key is not present, the default will be used. The keys are:
+
+            * standard_style. Create standard conflicted merge files.
+            * diff3_style. Create diff3-style file.
+            * simplify_alnum. Condense non-alphanumeric regions for simplified
+              diff file.
+            * ignore_whitespace. Ignore all whitespace.
+            * ignore_whitespace_change. Ignore changes in amount of whitespace.
+            * ignore_whitespace_eol. Ignore whitespace at end of line.
+            * patience. Use the "patience diff" algorithm
+            * minimal. Take extra time to find minimal diff
         """
 
         ancestor_ptr = ffi.new('git_tree **')
@@ -725,7 +868,7 @@ class BaseRepository(_Repository):
         ours = ours.peel(Tree)
         theirs = theirs.peel(Tree)
 
-        opts = self._merge_options(favor)
+        opts = self._merge_options(favor, flags, file_flags)
 
         ffi.buffer(ancestor_ptr)[:] = ancestor._pointer[:]
         ffi.buffer(ours_ptr)[:] = ours._pointer[:]
@@ -1181,7 +1324,8 @@ class BaseRepository(_Repository):
         return Index.from_c(self, cindex)
 
 
-class Branches(object):
+class Branches:
+
     def __init__(self, repository, flag=GIT_BRANCH_ALL, commit=None):
         self._repository = repository
         self._flag = flag
@@ -1244,7 +1388,8 @@ class Branches(object):
         return self.get(name) is not None
 
 
-class References(object):
+class References:
+
     def __init__(self, repository):
         self._repository = repository
 
